@@ -14,6 +14,7 @@ import copy
 
 from ruamel.yaml.compat import ordereddict, PY2, string_types, MutableSliceableSequence
 from ruamel.yaml.scalarstring import ScalarString
+from ruamel.yaml.anchor import Anchor
 
 if PY2:
     from collections import MutableSet, Sized, Set, MutableMapping, Mapping
@@ -21,18 +22,17 @@ else:
     from collections.abc import MutableSet, Sized, Set, MutableMapping, Mapping
 
 if False:  # MYPY
-    from typing import Any, Dict, Optional, List, Union, Optional  # NOQA
+    from typing import Any, Dict, Optional, List, Union, Optional, Iterator  # NOQA
 
 # fmt: off
-__all__ = ["CommentedSeq", "CommentedKeySeq",
-           "CommentedMap", "CommentedOrderedMap",
-           "CommentedSet", 'comment_attrib', 'merge_attrib']
+__all__ = ['CommentedSeq', 'CommentedKeySeq',
+           'CommentedMap', 'CommentedOrderedMap',
+           'CommentedSet', 'comment_attrib', 'merge_attrib']
 # fmt: on
 
 comment_attrib = '_yaml_comment'
 format_attrib = '_yaml_format'
 line_col_attrib = '_yaml_line_col'
-anchor_attrib = '_yaml_anchor'
 merge_attrib = '_yaml_merge'
 tag_attrib = '_yaml_tag'
 
@@ -163,16 +163,6 @@ class LineCol(object):
         self.data[key] = data
 
 
-class Anchor(object):
-    __slots__ = 'value', 'always_dump'
-    attrib = anchor_attrib
-
-    def __init__(self):
-        # type: () -> None
-        self.value = None
-        self.always_dump = False
-
-
 class Tag(object):
     """store tag information for roundtripping"""
 
@@ -385,16 +375,16 @@ class CommentedBase(object):
         raise NotImplementedError
 
 
-class CommentedSeq(MutableSliceableSequence, CommentedBase):
+class CommentedSeq(MutableSliceableSequence, list, CommentedBase):
     __slots__ = (Comment.attrib, '_lst')
 
     def __init__(self, *args, **kw):
         # type: (Any, Any) -> None
-        self._lst = list(*args, **kw)
+        list.__init__(self, *args, **kw)
 
     def __getsingleitem__(self, idx):
         # type: (Any) -> Any
-        return self._lst[idx]
+        return list.__getitem__(self, idx)
 
     def __setsingleitem__(self, idx, value):
         # type: (Any, Any) -> None
@@ -406,11 +396,11 @@ class CommentedSeq(MutableSliceableSequence, CommentedBase):
                 and isinstance(self[idx], ScalarString)
             ):
                 value = type(self[idx])(value)
-        self._lst.__setitem__(idx, value)
+        list.__setitem__(self, idx, value)
 
     def __delsingleitem__(self, idx=None):
         # type: (Any) -> Any
-        del self._lst[idx]
+        list.__delitem__(self, idx)
         self.ca.items.pop(idx, None)  # might not be there -> default value
         for list_index in sorted(self.ca.items):
             if list_index < idx:
@@ -419,12 +409,12 @@ class CommentedSeq(MutableSliceableSequence, CommentedBase):
 
     def __len__(self):
         # type: () -> int
-        return len(self._lst)
+        return list.__len__(self)
 
     def insert(self, idx, val):
         # type: (Any, Any) -> None
         """the comments after the insertion have to move forward"""
-        self._lst.insert(idx, val)
+        list.insert(self, idx, val)
         for list_index in sorted(self.ca.items, reverse=True):
             if list_index < idx:
                 break
@@ -432,11 +422,11 @@ class CommentedSeq(MutableSliceableSequence, CommentedBase):
 
     def extend(self, val):
         # type: (Any) -> None
-        self._lst.extend(val)
+        list.extend(self, val)
 
     def __eq__(self, other):
         # type: (Any) -> bool
-        return bool(self._lst == other)
+        return list.__eq__(self, other)
 
     def _yaml_add_comment(self, comment, key=NoComment):
         # type: (Any, Optional[Any]) -> None
@@ -493,11 +483,19 @@ class CommentedSeq(MutableSliceableSequence, CommentedBase):
         return res
 
     def __add__(self, other):
-        return self._lst + other
+        # type: (Any) -> Any
+        return list.__add__(self, other)
 
-    def sort(self):
-        tmp_lst = sorted(zip(self._lst, range(len(self._lst))))
-        self._lst = [x[0] for x in tmp_lst]
+    def sort(self, key=None, reverse=False):  # type: ignore
+        # type: (Any, bool) -> None
+        if key is None:
+            tmp_lst = sorted(zip(self, range(len(self))), reverse=reverse)
+            list.__init__(self, [x[0] for x in tmp_lst])
+        else:
+            tmp_lst = sorted(
+                zip(map(key, list.__iter__(self)), range(len(self))), reverse=reverse
+            )
+            list.__init__(self, [list.__getitem__(self, x[1]) for x in tmp_lst])
         itm = self.ca.items
         self.ca._items = {}
         for idx, x in enumerate(tmp_lst):
@@ -505,6 +503,10 @@ class CommentedSeq(MutableSliceableSequence, CommentedBase):
             if old_index in itm:
                 self.ca.items[idx] = itm[old_index]
 
+    def __repr__(self):
+        # type: () -> Any
+        return list.__repr__(self)
+
 
 class CommentedKeySeq(tuple, CommentedBase):
     """This primarily exists to be able to roundtrip keys that are sequences"""
@@ -627,12 +629,14 @@ class CommentedMapValuesView(CommentedMapView):
             yield self._mapping[key]
 
 
-class CommentedMap(CommentedBase, MutableMapping):
-    __slots__ = (Comment.attrib, '_od')
+class CommentedMap(MutableMapping, ordereddict, CommentedBase):
+    __slots__ = (Comment.attrib, '_ok', '_ref')
 
     def __init__(self, *args, **kw):
         # type: (Any, Any) -> None
-        self._od = ordereddict(*args, **kw)
+        self._ok = set()  # type: MutableSet[Any]  #  own keys
+        self._ref = []  # type: List[CommentedMap]
+        ordereddict.__init__(self, *args, **kw)  # type: ignore
 
     def _yaml_add_comment(self, comment, key=NoComment, value=NoComment):
         # type: (Any, Optional[Any], Optional[Any]) -> None
@@ -694,7 +698,8 @@ class CommentedMap(CommentedBase, MutableMapping):
     def update(self, vals):  # type: ignore
         # type: (Any) -> None
         try:
-            self._od.update(vals)
+            ordereddict.update(self, vals)
+            self._ok.update(vals.keys())  # type: ignore
         except TypeError:
             # probably a dict that is used
             for x in vals:
@@ -705,7 +710,8 @@ class CommentedMap(CommentedBase, MutableMapping):
         """insert key value into given position
         attach comment if provided
         """
-        self._od.insert(pos, key, value)
+        ordereddict.insert(self, pos, key, value)
+        self._ok.add(key)
         if comment is not None:
             self.yaml_add_eol_comment(comment, key=key)
 
@@ -738,7 +744,7 @@ class CommentedMap(CommentedBase, MutableMapping):
     def __getitem__(self, key):
         # type: (Any) -> Any
         try:
-            return self._od.__getitem__(key)
+            return ordereddict.__getitem__(self, key)
         except KeyError:
             for merged in getattr(self, merge_attrib, []):
                 if key in merged[1]:
@@ -755,23 +761,18 @@ class CommentedMap(CommentedBase, MutableMapping):
                 and isinstance(self[key], ScalarString)
             ):
                 value = type(self[key])(value)
-        self._od.__setitem__(key, value)
+        ordereddict.__setitem__(self, key, value)
+        self._ok.add(key)
 
     def _unmerged_contains(self, key):
         # type: (Any) -> Any
-        if self._od.__contains__(key):
+        if key in self._ok:
             return True
         return None
 
     def __contains__(self, key):
         # type: (Any) -> bool
-        if self._od.__contains__(key):
-            return True
-        # this will only work once the mapping/dict is built to completion
-        for merged in getattr(self, merge_attrib, []):
-            if key in merged[1]:
-                return True
-        return False
+        return bool(ordereddict.__contains__(self, key))
 
     def get(self, key, default=None):
         # type: (Any, Any) -> Any
@@ -782,77 +783,47 @@ class CommentedMap(CommentedBase, MutableMapping):
 
     def __repr__(self):
         # type: () -> Any
-        if not hasattr(self, merge_attrib):
-            return self._od.__repr__()
-        return 'ordereddict(' + repr(list(self._items())) + ')'
+        return ordereddict.__repr__(self).replace('CommentedMap', 'ordereddict')
 
     def non_merged_items(self):
         # type: () -> Any
-        for x in self._od.__iter__():
-            yield x, self._od.__getitem__(x)
+        for x in ordereddict.__iter__(self):
+            if x in self._ok:
+                yield x, ordereddict.__getitem__(self, x)
 
     def __delitem__(self, key):
         # type: (Any) -> None
-        found = True
-        for merged in getattr(self, merge_attrib, []):
-            try:
-                del merged[1][key]
-                found = True
-            except KeyError:
-                pass
-        try:
-            self._od.__delitem__(key)
-        except KeyError:
-            if not found:
-                raise
+        # for merged in getattr(self, merge_attrib, []):
+        #     if key in merged[1]:
+        #         value = merged[1][key]
+        #         break
+        # else:
+        #     # not found in merged in stuff
+        #     ordereddict.__delitem__(self, key)
+        #    for referer in self._ref:
+        #        referer.update_key_value(key)
+        #    return
+        #
+        # ordereddict.__setitem__(self, key, value)  # merge might have different value
+        # self._ok.discard(key)
+        self._ok.discard(key)
+        ordereddict.__delitem__(self, key)
+        for referer in self._ref:
+            referer.update_key_value(key)
 
     def __iter__(self):
         # type: () -> Any
-        for x in self._od.__iter__():
+        for x in ordereddict.__iter__(self):
             yield x
-        done = []  # type: List[Any]  # list of processed merge items, kept for masking
-        for merged in getattr(self, merge_attrib, []):
-            for x in merged[1]:
-                if self._od.__contains__(x):
-                    continue
-                for y in done:
-                    if x in y:
-                        break
-                else:
-                    yield x
-            done.append(merged[1])
 
     def _keys(self):
         # type: () -> Any
-        for x in self._od.__iter__():
+        for x in ordereddict.__iter__(self):
             yield x
-        done = []  # type: List[Any]  # list of processed merge items, kept for masking
-        for merged in getattr(self, merge_attrib, []):
-            for x in merged[1].keys():
-                if self._od.__contains__(x):
-                    continue
-                for y in done:
-                    if x in y:
-                        break
-                else:
-                    yield x
-            done.append(merged[1])
 
     def __len__(self):
         # type: () -> int
-        count = self._od.__len__()  # type: int
-        done = []  # type: List[Any] # list of processed merge items, kept for masking
-        for merged in getattr(self, merge_attrib, []):
-            for x in merged[1]:
-                if self._od.__contains__(x):
-                    continue
-                for y in done:
-                    if x in y:
-                        break
-                else:
-                    count += 1
-            done.append(merged[1])
-        return count
+        return ordereddict.__len__(self)  # type: ignore
 
     def __eq__(self, other):
         # type: (Any) -> bool
@@ -882,19 +853,8 @@ class CommentedMap(CommentedBase, MutableMapping):
 
         def _values(self):
             # type: () -> Any
-            for x in self._od.__iter__():
-                yield self._od.__getitem__(x)
-            done = []  # type: List[Any]  # list of processed merge items, kept for masking
-            for merged in getattr(self, merge_attrib, []):
-                for x in merged[1]:
-                    if self._od.__contains__(x):
-                        continue
-                    for y in done:
-                        if x in y:
-                            break
-                    else:
-                        yield merged[1]._od.__getitem__(x)
-                done.append(merged[1])
+            for x in ordereddict.__iter__(self):
+                yield ordereddict.__getitem__(self, x)
 
         def values(self):
             # type: () -> Any
@@ -916,19 +876,8 @@ class CommentedMap(CommentedBase, MutableMapping):
 
     def _items(self):
         # type: () -> Any
-        for x in self._od.__iter__():
-            yield x, self._od.__getitem__(x)
-        done = []  # type: List[Any]  # list of processed merge items, kept for masking
-        for merged in getattr(self, merge_attrib, []):
-            for x, v in merged[1].items():
-                if self._od.__contains__(x):
-                    continue
-                for y in done:
-                    if x in y:
-                        break
-                else:
-                    yield x, v  # self._od.__getitem__(merged[1], x)
-            done.append(merged[1])
+        for x in ordereddict.__iter__(self):
+            yield x, ordereddict.__getitem__(self, x)
 
     if PY2:
 
@@ -964,10 +913,31 @@ class CommentedMap(CommentedBase, MutableMapping):
             x[k] = v
         return x
 
+    def add_referent(self, cm):
+        # type: (Any) -> None
+        if cm not in self._ref:
+            self._ref.append(cm)
+
     def add_yaml_merge(self, value):
         # type: (Any) -> None
+        for v in value:
+            v[1].add_referent(self)
+            for k, v in v[1].items():
+                if ordereddict.__contains__(self, k):
+                    continue
+                ordereddict.__setitem__(self, k, v)
         self.merge.extend(value)
 
+    def update_key_value(self, key):
+        # type: (Any) -> None
+        if key in self._ok:
+            return
+        for v in self.merge:
+            if key in v[1]:
+                ordereddict.__setitem__(self, key, v[1][key])
+                return
+        ordereddict.__delitem__(self, key)
+
     def __deepcopy__(self, memo):
         # type: (Any) -> Any
         res = self.__class__()
@@ -979,8 +949,9 @@ class CommentedMap(CommentedBase, MutableMapping):
 
 
 # based on brownie mappings
-@classmethod
+@classmethod  # type: ignore
 def raise_immutable(cls, *args, **kwargs):
+    # type: (Any, *Any, **Any) -> None
     raise TypeError('{} objects are immutable'.format(cls.__name__))
 
 
@@ -1004,26 +975,31 @@ class CommentedKeyMap(CommentedBase, Mapping):
 
     # need to implement __getitem__, __iter__ and __len__
     def __getitem__(self, index):
+        # type: (Any) -> Any
         return self._od[index]
 
     def __iter__(self):
+        # type: () -> Iterator[Any]
         for x in self._od.__iter__():
             yield x
 
     def __len__(self):
+        # type: () -> int
         return len(self._od)
 
     def __hash__(self):
+        # type: () -> Any
         return hash(tuple(self.items()))
 
     def __repr__(self):
         # type: () -> Any
         if not hasattr(self, merge_attrib):
             return self._od.__repr__()
-        return 'ordereddict(' + repr(list(self._items())) + ')'
+        return 'ordereddict(' + repr(list(self._od.items())) + ')'
 
     @classmethod
     def fromkeys(keys, v=None):
+        # type: (Any, Any) -> Any
         return CommentedKeyMap(dict.fromkeys(keys, v))
 
     def _yaml_add_comment(self, comment, key=NoComment):
@@ -1150,14 +1126,14 @@ def dump_comments(d, name="", sep='.', out=sys.stdout):
     """
     if isinstance(d, dict) and hasattr(d, 'ca'):
         if name:
-            print(name)
-        print(d.ca, file=out)  # type: ignore
+            sys.stdout.write('{}\n'.format(name))
+        out.write('{}\n'.format(d.ca))  # type: ignore
         for k in d:
             dump_comments(d[k], name=(name + sep + k) if name else k, sep=sep, out=out)
     elif isinstance(d, list) and hasattr(d, 'ca'):
         if name:
-            print(name)
-        print(d.ca, file=out)  # type: ignore
+            sys.stdout.write('{}\n'.format(name))
+        out.write('{}\n'.format(d.ca))  # type: ignore
         for idx, k in enumerate(d):
             dump_comments(
                 k, name=(name + sep + str(idx)) if name else str(idx), sep=sep, out=out
